package sun.instrument;

import java.security.ProtectionDomain;
import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.lang.reflect.AccessibleObject;
import java.util.jar.JarFile;
import java.lang.instrument.ClassDefinition;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.Instrumentation;

public class InstrumentationImpl implements Instrumentation
{
    private final TransformerManager mTransformerManager;
    private TransformerManager mRetransfomableTransformerManager;
    private final long mNativeAgent;
    private final boolean mEnvironmentSupportsRedefineClasses;
    private volatile boolean mEnvironmentSupportsRetransformClassesKnown;
    private volatile boolean mEnvironmentSupportsRetransformClasses;
    private final boolean mEnvironmentSupportsNativeMethodPrefix;
    
    private InstrumentationImpl(long mNativeAgent, boolean mEnvironmentSupportsRedefineClasses, boolean mEnvironmentSupportsNativeMethodPrefix) {
        this.mTransformerManager = new TransformerManager(false);
        this.mRetransfomableTransformerManager = null;
        this.mNativeAgent = mNativeAgent;
        this.mEnvironmentSupportsRedefineClasses = mEnvironmentSupportsRedefineClasses;
        this.mEnvironmentSupportsRetransformClassesKnown = false;
        this.mEnvironmentSupportsRetransformClasses = false;
        this.mEnvironmentSupportsNativeMethodPrefix = mEnvironmentSupportsNativeMethodPrefix;
    }
    
    @Override
    public void addTransformer(ClassFileTransformer classFileTransformer) {
        this.addTransformer(classFileTransformer, false);
    }
    
    @Override
    public synchronized void addTransformer(ClassFileTransformer classFileTransformer, boolean b) {
        if (classFileTransformer == null) {
            throw new NullPointerException("null passed as 'transformer' in addTransformer");
        }
        if (b) {
            if (!this.isRetransformClassesSupported()) {
                throw new UnsupportedOperationException("adding retransformable transformers is not supported in this environment");
            }
            if (this.mRetransfomableTransformerManager == null) {
                this.mRetransfomableTransformerManager = new TransformerManager(true);
            }
            this.mRetransfomableTransformerManager.addTransformer(classFileTransformer);
            if (this.mRetransfomableTransformerManager.getTransformerCount() == 1) {
                this.setHasRetransformableTransformers(this.mNativeAgent, true);
            }
        }
        else {
            this.mTransformerManager.addTransformer(classFileTransformer);
        }
        return;
    }
    
    @Override
    public synchronized boolean removeTransformer(ClassFileTransformer classFileTransformer) {
        if (classFileTransformer == null) {
            throw new NullPointerException("null passed as 'transformer' in removeTransformer");
        }
        TransformerManager transformerManager = this.findTransformerManager(classFileTransformer);
        if (transformerManager != null) {
            transformerManager.removeTransformer(classFileTransformer);
            if (transformerManager.isRetransformable() && transformerManager.getTransformerCount() == 0) {
                this.setHasRetransformableTransformers(this.mNativeAgent, false);
            }
            return true;
        }
        else {
            return false;
        }
    }
    
    @Override
    public boolean isModifiableClass(Class<?> clazz) {
        if (clazz == null) {
            throw new NullPointerException("null passed as 'theClass' in isModifiableClass");
        }
        return this.isModifiableClass0(this.mNativeAgent, clazz);
    }
    
    @Override
    public boolean isRetransformClassesSupported() {
        if (!this.mEnvironmentSupportsRetransformClassesKnown) {
            this.mEnvironmentSupportsRetransformClasses = this.isRetransformClassesSupported0(this.mNativeAgent);
            this.mEnvironmentSupportsRetransformClassesKnown = true;
        }
        return this.mEnvironmentSupportsRetransformClasses;
    }
    
    @Override
    public void retransformClasses(Class<?>... array) {
        if (!this.isRetransformClassesSupported()) {
            throw new UnsupportedOperationException("retransformClasses is not supported in this environment");
        }
        this.retransformClasses0(this.mNativeAgent, array);
        return;
    }
    
    @Override
    public boolean isRedefineClassesSupported() {
        return this.mEnvironmentSupportsRedefineClasses;
    }
    
    @Override
    public void redefineClasses(ClassDefinition... array) throws ClassNotFoundException {
        if (!this.isRedefineClassesSupported()) {
            throw new UnsupportedOperationException("redefineClasses is not supported in this environment");
        }
        if (array == null) {
            throw new NullPointerException("null passed as 'definitions' in redefineClasses");
        }
        int i = 0;
        while (i < array.length) {
            if (array[i] == null) {
                throw new NullPointerException("element of 'definitions' is null in redefineClasses");
            }
            ++i;
        }
        if (array.length == 0) {
            return;
        }
        this.redefineClasses0(this.mNativeAgent, array);
        return;
    }
    
    @Override
    public Class[] getAllLoadedClasses() {
        return this.getAllLoadedClasses0(this.mNativeAgent);
    }
    
    @Override
    public Class[] getInitiatedClasses(ClassLoader classLoader) {
        return this.getInitiatedClasses0(this.mNativeAgent, classLoader);
    }
    
    @Override
    public long getObjectSize(Object o) {
        if (o == null) {
            throw new NullPointerException("null passed as 'objectToSize' in getObjectSize");
        }
        return this.getObjectSize0(this.mNativeAgent, o);
    }
    
    @Override
    public void appendToBootstrapClassLoaderSearch(JarFile jarFile) {
        this.appendToClassLoaderSearch0(this.mNativeAgent, jarFile.getName(), true);
    }
    
    @Override
    public void appendToSystemClassLoaderSearch(JarFile jarFile) {
        this.appendToClassLoaderSearch0(this.mNativeAgent, jarFile.getName(), false);
    }
    
    @Override
    public boolean isNativeMethodPrefixSupported() {
        return this.mEnvironmentSupportsNativeMethodPrefix;
    }
    
    @Override
    public synchronized void setNativeMethodPrefix(ClassFileTransformer classFileTransformer, String s) {
        if (!this.isNativeMethodPrefixSupported()) {
            throw new UnsupportedOperationException("setNativeMethodPrefix is not supported in this environment");
        }
        if (classFileTransformer == null) {
            throw new NullPointerException("null passed as 'transformer' in setNativeMethodPrefix");
        }
        TransformerManager transformerManager = this.findTransformerManager(classFileTransformer);
        if (transformerManager == null) {
            throw new IllegalArgumentException("transformer not registered in setNativeMethodPrefix");
        }
        transformerManager.setNativeMethodPrefix(classFileTransformer, s);
        this.setNativeMethodPrefixes(this.mNativeAgent, transformerManager.getNativeMethodPrefixes(), transformerManager.isRetransformable());
        return;
    }
    
    private TransformerManager findTransformerManager(ClassFileTransformer classFileTransformer) {
        if (this.mTransformerManager.includesTransformer(classFileTransformer)) {
            return this.mTransformerManager;
        }
        if (this.mRetransfomableTransformerManager != null && this.mRetransfomableTransformerManager.includesTransformer(classFileTransformer)) {
            return this.mRetransfomableTransformerManager;
        }
        return null;
    }
    
    private native boolean isModifiableClass0(long p0, Class<?> p1);
    
    private native boolean isRetransformClassesSupported0(long p0);
    
    private native void setHasRetransformableTransformers(long p0, boolean p1);
    
    private native void retransformClasses0(long p0, Class<?>[] p1);
    
    private native void redefineClasses0(long p0, ClassDefinition[] p1) throws ClassNotFoundException;
    
    private native Class[] getAllLoadedClasses0(long p0);
    
    private native Class[] getInitiatedClasses0(long p0, ClassLoader p1);
    
    private native long getObjectSize0(long p0, Object p1);
    
    private native void appendToClassLoaderSearch0(long p0, String p1, boolean p2);
    
    private native void setNativeMethodPrefixes(long p0, String[] p1, boolean p2);
    
    private static void setAccessible(final AccessibleObject accessibleObject, final boolean b) {
        AccessController.doPrivileged(new PrivilegedAction<Object>() {
            @Override
            public Object run() {
                accessibleObject.setAccessible(b);
                return null;
            }
        });
    }
    
    private void loadClassAndStartAgent(String name, String s, String s2) throws Throwable {
        Class<Object> loadClass = ClassLoader.getSystemClassLoader().loadClass(name);
        Method method = null;
        NoSuchMethodException ex = null;
        boolean n = false;
        try {
            method = loadClass.getDeclaredMethod(s, new Class[] { String.class, Instrumentation.class });
            n = true;
        }
        catch (NoSuchMethodException ex2) {
            ex = ex2;
        }
        if (method == null) {
            try {
                method = loadClass.getDeclaredMethod(s, new Class[] { String.class });
            }
            catch (NoSuchMethodException ex3) {}
        }
        if (method == null) {
            try {
                method = loadClass.getMethod(s, new Class[] { String.class, Instrumentation.class });
                n = true;
            }
            catch (NoSuchMethodException ex4) {}
        }
        if (method == null) {
            try {
                method = loadClass.getMethod(s, new Class[] { String.class });
            }
            catch (NoSuchMethodException ex5) {
                throw ex;
            }
        }
        setAccessible(method, true);
        if (!(n == 0)) {
            method.invoke(null, new Object[] { s2, this });
        }
        else {
            method.invoke(null, new Object[] { s2 });
        }
        setAccessible(method, false);
    }
    
    private void loadClassAndCallPremain(String s, String s2) throws Throwable {
        this.loadClassAndStartAgent(s, "premain", s2);
    }
    
    private void loadClassAndCallAgentmain(String s, String s2) throws Throwable {
        this.loadClassAndStartAgent(s, "agentmain", s2);
    }
    
    private byte[] transform(ClassLoader classLoader, String s, Class<?> clazz, ProtectionDomain protectionDomain, byte[] array, boolean b) {
        TransformerManager transformerManager = b ? this.mRetransfomableTransformerManager : this.mTransformerManager;
        if (transformerManager == null) {
            return null;
        }
        return transformerManager.transform(classLoader, s, clazz, protectionDomain, array);
    }
    
    static {
        System.loadLibrary("instrument");
    }
}
